# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import os
import tempfile
import binascii
import unittest.mock as mock
import mimetypes
from pathlib import Path

import botocore
import kserve
import pytest

STORAGE_MODULE = 'kserve.storage'
HTTPS_URI_TARGZ = 'https://foo.bar/model.tar.gz'
HTTPS_URI_TARGZ_WITH_QUERY = HTTPS_URI_TARGZ + '?foo=bar'

# *.tar.gz contains a single empty file model.pth
FILE_TAR_GZ_RAW = binascii.unhexlify('1f8b0800bac550600003cbcd4f49cdd12b28c960a01d3030303033315100d1e666a660dac008c28'
                                     '701054313a090a189919981998281a1b1b1a1118382010ddd0407a5c525894540a754656466e464e'
                                     '2560754969686c71ca83fe0f4281805a360140c7200009f7e1bb400060000')
# *.zip contains a single empty file model.pth
FILE_ZIP_RAW = binascii.unhexlify('504b030414000800080035b67052000000000000000000000000090020006d6f64656c2e70746855540'
                                  'd000786c5506086c5506086c5506075780b000104f501000004140000000300504b07080000000002000'
                                  '00000000000504b0102140314000800080035b6705200000000020000000000000009002000000000000'
                                  '0000000a481000000006d6f64656c2e70746855540d000786c5506086c5506086c5506075780b000104f'
                                  '50100000414000000504b0506000000000100010057000000590000000000')


def test_storage_local_path():
    abs_path = 'file:///'
    relative_path = 'file://.'
    assert kserve.Storage.download(abs_path) == abs_path.replace("file://", "", 1)
    assert kserve.Storage.download(relative_path) == relative_path.replace("file://", "", 1)


def test_storage_local_path_exception():
    not_exist_path = 'file:///some/random/path'
    with pytest.raises(Exception):
        kserve.Storage.download(not_exist_path)


def test_no_prefix_local_path():
    abs_path = '/'
    relative_path = '.'
    assert kserve.Storage.download(abs_path) == abs_path
    assert kserve.Storage.download(relative_path) == relative_path


class MockHttpResponse(object):
    def __init__(
        self,
        status_code=404,
        raw=b'',
        content_type=''
    ):
        self.status_code = status_code
        self.raw = io.BytesIO(raw)
        self.headers = {'Content-Type': content_type}

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_val, traceback):
        pass


@mock.patch('requests.get', return_value=MockHttpResponse(status_code=200, content_type='application/octet-stream'))
def test_http_uri_path(_):
    http_uri = 'http://foo.bar/model.joblib'
    http_with_query_uri = 'http://foo.bar/model.joblib?foo=bar'
    out_dir = '.'
    assert kserve.Storage.download(http_uri, out_dir=out_dir) == out_dir
    assert kserve.Storage.download(http_with_query_uri, out_dir=out_dir) == out_dir
    os.remove('./model.joblib')


@mock.patch('requests.get', return_value=MockHttpResponse(status_code=200, content_type='application/octet-stream'))
def test_https_uri_path(_):
    https_uri = 'https://foo.bar/model.joblib'
    https_with_query_uri = 'https://foo.bar/model.joblib?foo=bar'
    out_dir = '.'
    assert kserve.Storage.download(https_uri, out_dir=out_dir) == out_dir
    assert kserve.Storage.download(https_with_query_uri, out_dir=out_dir) == out_dir
    os.remove('./model.joblib')


http_uri_path_testparams = [
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-tar'), None),
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-gtar'), None),
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-gzip'), None),
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/gzip'), None),
    (HTTPS_URI_TARGZ, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/zip'), RuntimeError),
    (HTTPS_URI_TARGZ_WITH_QUERY, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-tar'), None),
    (HTTPS_URI_TARGZ_WITH_QUERY, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-gtar'), None),
    (HTTPS_URI_TARGZ_WITH_QUERY, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/x-gzip'), None),
    (HTTPS_URI_TARGZ_WITH_QUERY, MockHttpResponse(200, FILE_TAR_GZ_RAW, 'application/gzip'), None),
    ('https://foo.bar/model.zip', MockHttpResponse(200, FILE_ZIP_RAW, 'application/zip'), None),
    ('https://foo.bar/model.zip', MockHttpResponse(200, FILE_ZIP_RAW, 'application/x-zip-compressed'), None),
    ('https://foo.bar/model.zip', MockHttpResponse(200, FILE_ZIP_RAW, 'application/zip-compressed'), None),
    ('https://foo.bar/model.zip?foo=bar', MockHttpResponse(200, FILE_ZIP_RAW, 'application/zip'), None),
    ('https://foo.bar/model.zip?foo=bar', MockHttpResponse(200, FILE_ZIP_RAW, 'application/x-zip-compressed'), None),
    ('https://foo.bar/model.zip?foo=bar', MockHttpResponse(200, FILE_ZIP_RAW, 'application/zip-compressed'), None),
    ('https://theabyss.net/model.joblib', MockHttpResponse(404), RuntimeError),
    ('https://some.site.com/test.model', MockHttpResponse(status_code=200, content_type='text/html'), RuntimeError),
    ('https://foo.bar/test/', MockHttpResponse(200), ValueError),
]


@pytest.mark.parametrize('uri,response,expected_error', http_uri_path_testparams)
def test_http_uri_paths(uri, response, expected_error):
    if expected_error:
        def test(_):
            with pytest.raises(expected_error):
                kserve.Storage.download(uri)
    else:
        def test(_):
            with tempfile.TemporaryDirectory() as out_dir:
                assert kserve.Storage.download(uri, out_dir=out_dir) == out_dir
                assert os.path.exists(os.path.join(out_dir, 'model.pth'))
    mock.patch('requests.get', return_value=response)(test)()


@mock.patch(STORAGE_MODULE + '.storage')
def test_mock_gcs(mock_storage):
    gcs_path = 'gs://foo/bar'
    mock_obj = mock.MagicMock()
    mock_obj.name = 'mock.object'
    mock_storage.Client().bucket().list_blobs().__iter__.return_value = [mock_obj]
    assert kserve.Storage.download(gcs_path)


def test_storage_blob_exception():
    blob_path = 'https://accountname.blob.core.windows.net/container/some/blob/'
    with pytest.raises(Exception):
        kserve.Storage.download(blob_path)


@mock.patch(STORAGE_MODULE + '.boto3')
def test_storage_s3_exception(mock_boto3):
    path = 's3://foo/bar'
    # Create mock client
    mock_s3_resource = mock.MagicMock()
    mock_s3_resource.Bucket.side_effect = Exception()
    mock_boto3.resource.return_value = mock_s3_resource

    with pytest.raises(Exception):
        kserve.Storage.download(path)


@mock.patch(STORAGE_MODULE + '.boto3')
@mock.patch('urllib3.PoolManager')
def test_no_permission_buckets(mock_connection, mock_boto3):
    bad_s3_path = "s3://random/path"
    # Access private buckets without credentials
    mock_s3_resource = mock.MagicMock()
    mock_s3_bucket = mock.MagicMock()
    mock_s3_bucket.objects.filter.return_value = [mock.MagicMock()]
    mock_s3_bucket.objects.filter.side_effect = botocore.exceptions.ClientError(
        {}, "GetObject"
    )
    mock_s3_resource.Bucket.return_value = mock_s3_bucket
    mock_boto3.resource.return_value = mock_s3_resource

    with pytest.raises(botocore.exceptions.ClientError):
        kserve.Storage.download(bad_s3_path)


def test_unpack_tar_file():
    out_dir = '.'
    tar_file = os.path.join(out_dir, "model.tgz")
    Path(tar_file).write_bytes(FILE_TAR_GZ_RAW)
    mimetype, _ = mimetypes.guess_type(tar_file)
    kserve.Storage._unpack_archive_file(tar_file, mimetype, out_dir)
    assert os.path.exists(os.path.join(out_dir, 'model.pth'))
    os.remove(os.path.join(out_dir, 'model.pth'))


def test_unpack_zip_file():
    out_dir = '.'
    tar_file = os.path.join(out_dir, "model.zip")
    Path(tar_file).write_bytes(FILE_ZIP_RAW)
    mimetype, _ = mimetypes.guess_type(tar_file)
    kserve.Storage._unpack_archive_file(tar_file, mimetype, out_dir)
    assert os.path.exists(os.path.join(out_dir, 'model.pth'))
    os.remove(os.path.join(out_dir, 'model.pth'))



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import io
import json
import os
import re
from typing import Dict
from unittest import mock

import avro.io
import avro.schema
import pytest
from cloudevents.conversion import to_binary, to_structured
from cloudevents.http import CloudEvent
from fastapi.testclient import TestClient
from ray import serve

from kserve import Model, ModelServer, ModelRepository
from kserve.errors import InvalidInput
from kserve.model import PredictorProtocol
from kserve.protocol.rest.server import RESTServer

test_avsc_schema = '''
        {
        "namespace": "example.avro",
         "type": "record",
         "name": "User",
         "fields": [
             {"name": "name", "type": "string"},
             {"name": "favorite_number",  "type": ["int", "null"]},
             {"name": "favorite_color", "type": ["string", "null"]}
         ]
        }
        '''


def dummy_cloud_event(data, set_contenttype=False, add_extension=False):
    # This data defines a binary cloudevent
    attributes = {
        "type": "com.example.sampletype1",
        "source": "https://example.com/event-producer",
        "specversion": "1.0",
        "id": "36077800-0c23-4f38-a0b4-01f4369f670a",
        "time": "2021-01-28T21:04:43.144141+00:00"
    }
    if set_contenttype:
        attributes["content-type"] = "application/json"
    if add_extension:
        attributes["custom-extension"] = "custom-value"

    event = CloudEvent(attributes, data)
    return event


class DummyModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        return {"predictions": request["instances"]}


@serve.deployment
class DummyServeModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        return {"predictions": request["instances"]}


class DummyCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    async def predict(self, request, headers=None):
        return {"predictions": request["instances"]}

    async def explain(self, request, headers=None):
        return {"predictions": request["instances"]}


class DummyAvroCEModel(Model):
    def __init__(self, name):
        super().__init__(name)
        self.name = name
        self.ready = False

    def load(self):
        self.ready = True

    def _parserequest(self, request):
        schema = avro.schema.parse(test_avsc_schema)
        raw_bytes = request
        bytes_reader = io.BytesIO(raw_bytes)
        decoder = avro.io.BinaryDecoder(bytes_reader)
        reader = avro.io.DatumReader(schema)
        record1 = reader.read(decoder)
        return record1

    def preprocess(self, request, headers: Dict[str, str] = None):
        if isinstance(request, CloudEvent):
            attributes = request._attributes
            assert attributes["specversion"] == "1.0"
            assert attributes["source"] == "https://example.com/event-producer"
            assert attributes["type"] == "com.example.sampletype1"
            assert attributes["content-type"] == "application/json"
            return self._parserequest(request.data)

    async def predict(self, request, headers=None):
        return {"predictions": [[request['name'], request['favorite_number'], request['favorite_color']]]}

    async def explain(self, request, headers=None):
        return {"predictions": [[request['name'], request['favorite_number'], request['favorite_color']]]}


class DummyModelRepository(ModelRepository):
    def __init__(self, test_load_success: bool, fail_with_exception: bool = False):
        super().__init__()
        self.test_load_success = test_load_success
        self.fail_with_exception = fail_with_exception

    async def load(self, name: str) -> bool:
        if self.test_load_success:
            model = DummyModel(name)
            model.load()
            self.update(model)
            return model.ready
        else:
            if self.fail_with_exception:
                raise Exception(f"Could not load model {name}.")
            else:
                return False


@pytest.mark.asyncio
class TestModel:

    async def test_validate(self):
        model = DummyModel("TestModel")
        good_request = {"instances": []}
        validated_request = model.validate(good_request)
        assert validated_request == good_request
        bad_request = {"instances": "invalid"}
        with pytest.raises(InvalidInput):
            model.validate(bad_request)

        model.protocol = PredictorProtocol.REST_V2.value
        good_request = {"inputs": []}
        validated_request = model.validate(good_request)
        assert validated_request == good_request
        bad_request = {"inputs": "invalid"}
        with pytest.raises(InvalidInput):
            model.validate(bad_request)


class TestTFHttpServer:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        model.load()
        server = ModelServer()
        server.register_model(model)
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_liveness(self, http_server_client):
        resp = http_server_client.get('/')
        assert resp.status_code == 200
        assert resp.json() == {"status": "alive"}

    def test_model(self, http_server_client):
        resp = http_server_client.get('/v1/models/TestModel')
        assert resp.status_code == 200

    def test_unknown_model(self, http_server_client):
        resp = http_server_client.get('/v1/models/InvalidModel')
        assert resp.status_code == 404
        assert resp.json() == {"error": "Model with name InvalidModel does not exist."}

    def test_list_models(self, http_server_client):
        resp = http_server_client.get('/v1/models')
        assert resp.status_code == 200
        assert resp.json() == {"models": ["TestModel"]}

    def test_predict(self, http_server_client):
        resp = http_server_client.post('/v1/models/TestModel:predict',
                                       data=b'{"instances":[[1,2]]}')
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers['content-type'] == "application/json"

    def test_explain(self, http_server_client):
        resp = http_server_client.post('/v1/models/TestModel:explain',
                                       data=b'{"instances":[[1,2]]}')
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers['content-type'] == "application/json"

    def test_unknown_path(self, http_server_client):
        resp = http_server_client.get('/unknown_path')
        assert resp.status_code == 404
        assert resp.json() == {"detail": "Not Found"}

    def test_metrics(self, http_server_client):
        resp = http_server_client.get('/metrics')
        assert resp.status_code == 200
        assert resp.content is not None


class TestRayServer:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        serve.start(detached=False, http_options={"host": "0.0.0.0", "port": 9071})

        DummyServeModel.deploy("TestModel")
        handle = DummyServeModel.get_handle()
        handle.load.remote()

        server = ModelServer()
        server.register_model_handle("TestModel", handle)
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_liveness_handler(self, http_server_client):
        resp = http_server_client.get('/')
        assert resp.status_code == 200
        assert resp.content == b'{"status":"alive"}'

    def test_list_handler(self, http_server_client):
        resp = http_server_client.get('/v1/models')
        assert resp.status_code == 200
        assert resp.content == b'{"models":["TestModel"]}'

    def test_health_handler(self, http_server_client):
        resp = http_server_client.get('/v1/models/TestModel')
        assert resp.status_code == 200
        assert resp.content == b'{"name":"TestModel","ready":true}'

    def test_predict(self, http_server_client):
        resp = http_server_client.post('/v1/models/TestModel:predict',
                                       data=b'{"instances":[[1,2]]}')
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers['content-type'] == "application/json"

    def test_explain(self, http_server_client):
        resp = http_server_client.post('/v1/models/TestModel:explain',
                                       data=b'{"instances":[[1,2]]}')
        assert resp.status_code == 200
        assert resp.content == b'{"predictions":[[1,2]]}'
        assert resp.headers['content-type'] == "application/json"


class TestTFHttpServerModelNotLoaded:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_model_not_ready_error(self, http_server_client):
        resp = http_server_client.get('/v1/models/TestModel')
        assert resp.status_code == 503


class TestTFHttpServerCloudEvent:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyCEModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_predict_ce_structured(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]})
        headers, body = to_structured(event)

        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)
        body = json.loads(resp.content)

        assert resp.status_code == 200
        assert resp.headers['content-type'] == "application/cloudevents+json"

        assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert body["data"] == {"predictions": [[1, 2]]}
        assert body['specversion'] == "1.0"
        assert body['source'] == "io.kserve.inference.TestModel"
        assert body['type'] == "io.kserve.inference.response"
        assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    def test_predict_custom_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ,
                             {"CE_SOURCE": "io.kserve.inference.CustomSource", "CE_TYPE": "io.kserve.custom_type"}):
            event = dummy_cloud_event({"instances": [[1, 2]]})
            headers, body = to_structured(event)

            resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)
            body = json.loads(resp.content)

            assert resp.status_code == 200
            assert resp.headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.inference.CustomSource"
            assert body['type'] == "io.kserve.custom_type"

    def test_predict_merge_structured_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, add_extension=True)
            headers, body = to_structured(event)

            resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)
            body = json.loads(resp.content)

            assert resp.status_code == 200
            assert resp.headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.inference.TestModel"
            assert body['type'] == "io.kserve.inference.response"
            assert body["custom-extension"] == "custom-value"  # Added by add_extension=True in dummy_cloud_event
            assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    def test_predict_merge_binary_ce_attributes(self, http_server_client):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True, add_extension=True)
            headers, body = to_binary(event)

            resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

            assert resp.status_code == 200
            assert resp.headers['content-type'] == "application/json"
            assert resp.headers['ce-specversion'] == "1.0"
            assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            # Added by add_extension=True in dummy_cloud_event
            assert resp.headers['ce-custom-extension'] == 'custom-value'
            assert resp.headers['ce-source'] == "io.kserve.inference.TestModel"
            assert resp.headers['ce-type'] == "io.kserve.inference.response"
            assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
            assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_binary_dict(self, http_server_client):
        event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

        assert resp.status_code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.inference.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_binary_bytes(self, http_server_client):
        event = dummy_cloud_event(b'{"instances":[[1,2]]}', set_contenttype=True)
        headers, body = to_binary(event)
        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

        assert resp.status_code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.inference.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [[1, 2]]}'

    def test_predict_ce_bytes_bad_format_exception(self, http_server_client):
        event = dummy_cloud_event(b'{', set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

        assert resp.status_code == 400
        error_regex = re.compile("Failed to decode or parse binary json cloudevent: "
                                 "unexpected end of data:*")
        response = json.loads(resp.content)
        assert error_regex.match(response["error"]) is not None

    def test_predict_ce_bytes_bad_hex_format_exception(self, http_server_client):
        event = dummy_cloud_event(b'0\x80\x80\x06World!\x00\x00', set_contenttype=True)
        headers, body = to_binary(event)

        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

        assert resp.status_code == 400
        error_regex = re.compile("Failed to decode or parse binary json cloudevent: "
                                 "'utf-8' codec can't decode byte 0x80 in position 1: invalid start byte.*")
        response = json.loads(resp.content)
        assert error_regex.match(response["error"]) is not None


class TestTFHttpServerAvroCloudEvent:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        model = DummyAvroCEModel("TestModel")
        server = ModelServer()
        server.register_model(model)
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_predict_ce_avro_binary(self, http_server_client):
        schema = avro.schema.parse(test_avsc_schema)
        msg = {"name": "foo", "favorite_number": 1, "favorite_color": "pink"}

        writer = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer.write(msg, encoder)
        data = bytes_writer.getvalue()

        event = dummy_cloud_event(data, set_contenttype=True)
        # Creates the HTTP request representation of the CloudEvent in binary content mode
        headers, body = to_binary(event)
        resp = http_server_client.post('/v1/models/TestModel:predict', headers=headers, data=body)

        assert resp.status_code == 200
        assert resp.headers['content-type'] == "application/json"
        assert resp.headers['ce-specversion'] == "1.0"
        assert resp.headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert resp.headers['ce-source'] == "io.kserve.inference.TestModel"
        assert resp.headers['ce-type'] == "io.kserve.inference.response"
        assert resp.headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert resp.content == b'{"predictions": [["foo", 1, "pink"]]}'


class TestTFHttpServerLoadAndUnLoad:

    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        server = ModelServer(registered_models=DummyModelRepository(test_load_success=True))
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_load(self, http_server_client):
        resp = http_server_client.post('/v2/repository/models/model/load', data=b'')
        assert resp.status_code == 200
        assert resp.content == b'{"name":"model","load":true}'

    def test_unload(self, http_server_client):
        resp = http_server_client.post('/v2/repository/models/model/unload', data=b'')
        assert resp.status_code == 200
        assert resp.content == b'{"name":"model","unload":true}'


class TestTFHttpServerLoadAndUnLoadFailure:
    @pytest.fixture(scope="class")
    def app(self):  # pylint: disable=no-self-use
        server = ModelServer(registered_models=DummyModelRepository(test_load_success=False))
        rest_server = RESTServer(server.dataplane, server.model_repository_extension)
        return rest_server.create_application()

    @pytest.fixture(scope='class')
    def http_server_client(self, app):
        return TestClient(app)

    def test_load_fail(self, http_server_client):
        resp = http_server_client.post('/v2/repository/models/model/load', data=b'')
        assert resp.status_code == 503

    def test_unload_fail(self, http_server_client):
        resp = http_server_client.post('/v2/repository/models/model/unload', data=b'')
        assert resp.status_code == 404



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from prometheus_client import Histogram

PROM_LABELS = ['model_name']
PRE_HIST_TIME = Histogram('request_preprocess_seconds', 'pre-process request latency', PROM_LABELS)
POST_HIST_TIME = Histogram('request_postprocess_seconds', 'post-process request latency', PROM_LABELS)
PREDICT_HIST_TIME = Histogram('request_predict_seconds', 'predict request latency', PROM_LABELS)
EXPLAIN_HIST_TIME = Histogram('request_explain_seconds', 'explain request latency', PROM_LABELS)


def get_labels(model_name):
    return {PROM_LABELS[0]: model_name}



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service_spec import V1beta1InferenceServiceSpec  # noqa: E501
from kserve.rest import ApiException

class TestV1beta1InferenceServiceSpec(unittest.TestCase):
    """V1beta1InferenceServiceSpec unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceServiceSpec
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # model = kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec()  # noqa: E501
        if include_optional :
            return V1beta1InferenceServiceSpec(
                explainer = kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                    active_deadline_seconds = 56, 
                    affinity = None, 
                    aix = kserve.models.v1beta1_aix_explainer_spec.V1beta1AIXExplainerSpec(
                        args = [
                            '0'
                            ], 
                        command = [
                            '0'
                            ], 
                        config = {
                            'key' : '0'
                            }, 
                        env = [
                            None
                            ], 
                        env_from = [
                            None
                            ], 
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        ports = [
                            None
                            ], 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        type = '0', 
                        volume_devices = [
                            None
                            ], 
                        volume_mounts = [
                            None
                            ], 
                        working_dir = '0', ), 
                    alibi = kserve.models.v1beta1_alibi_explainer_spec.V1beta1AlibiExplainerSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        type = '0', 
                        working_dir = '0', ), 
                    automount_service_account_token = True, 
                    batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size = 56, 
                        max_latency = 56, 
                        timeout = 56, ), 
                    canary_traffic_percent = 56, 
                    container_concurrency = 56, 
                    containers = [
                        None
                        ], 
                    dns_config = None, 
                    dns_policy = '0', 
                    enable_service_links = True, 
                    ephemeral_containers = [
                        None
                        ], 
                    host_aliases = [
                        None
                        ], 
                    host_ipc = True, 
                    host_network = True, 
                    host_pid = True, 
                    hostname = '0', 
                    image_pull_secrets = [
                        None
                        ], 
                    init_containers = [
                        None
                        ], 
                    logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode = '0', 
                        url = '0', ), 
                    max_replicas = 56, 
                    min_replicas = 56, 
                    node_name = '0', 
                    node_selector = {
                        'key' : '0'
                        }, 
                    overhead = {
                        'key' : None
                        }, 
                    preemption_policy = '0', 
                    priority = 56, 
                    priority_class_name = '0', 
                    readiness_gates = [
                        None
                        ], 
                    restart_policy = '0', 
                    runtime_class_name = '0', 
                    scheduler_name = '0', 
                    security_context = None, 
                    service_account = '0', 
                    service_account_name = '0', 
                    share_process_namespace = True, 
                    subdomain = '0', 
                    termination_grace_period_seconds = 56, 
                    timeout = 56, 
                    tolerations = [
                        None
                        ], 
                    topology_spread_constraints = [
                        None
                        ], 
                    volumes = [
                        None
                        ], ), 
                predictor = kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                    active_deadline_seconds = 56, 
                    affinity = None, 
                    automount_service_account_token = True, 
                    batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size = 56, 
                        max_latency = 56, 
                        timeout = 56, ), 
                    canary_traffic_percent = 56, 
                    container_concurrency = 56, 
                    containers = [
                        None
                        ], 
                    dns_config = None, 
                    dns_policy = '0', 
                    enable_service_links = True, 
                    ephemeral_containers = [
                        None
                        ], 
                    host_aliases = [
                        None
                        ], 
                    host_ipc = True, 
                    host_network = True, 
                    host_pid = True, 
                    hostname = '0', 
                    image_pull_secrets = [
                        None
                        ], 
                    init_containers = [
                        None
                        ], 
                    logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode = '0', 
                        url = '0', ), 
                    max_replicas = 56, 
                    min_replicas = 56, 
                    node_name = '0', 
                    node_selector = {
                        'key' : '0'
                        }, 
                    onnx = kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                        args = [
                            '0'
                            ], 
                        command = [
                            '0'
                            ], 
                        env = [
                            None
                            ], 
                        env_from = [
                            None
                            ], 
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        ports = [
                            None
                            ], 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        volume_devices = [
                            None
                            ], 
                        volume_mounts = [
                            None
                            ], 
                        working_dir = '0', ), 
                    overhead = {
                        'key' : None
                        }, 
                    preemption_policy = '0', 
                    priority = 56, 
                    priority_class_name = '0', 
                    pytorch = kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0',
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    readiness_gates = [
                        None
                        ], 
                    restart_policy = '0', 
                    runtime_class_name = '0', 
                    scheduler_name = '0', 
                    security_context = None, 
                    service_account = '0', 
                    service_account_name = '0', 
                    share_process_namespace = True, 
                    sklearn = kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    subdomain = '0', 
                    tensorflow = kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    termination_grace_period_seconds = 56, 
                    timeout = 56, 
                    tolerations = [
                        None
                        ], 
                    topology_spread_constraints = [
                        None
                        ], 
                    triton = kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    volumes = [
                        None
                        ], 
                    xgboost = kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), ), 
                transformer = kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                    active_deadline_seconds = 56, 
                    affinity = None, 
                    automount_service_account_token = True, 
                    batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size = 56, 
                        max_latency = 56, 
                        timeout = 56, ), 
                    canary_traffic_percent = 56, 
                    container_concurrency = 56, 
                    containers = [
                        None
                        ], 
                    dns_config = None, 
                    dns_policy = '0', 
                    enable_service_links = True, 
                    ephemeral_containers = [
                        None
                        ], 
                    host_aliases = [
                        None
                        ], 
                    host_ipc = True, 
                    host_network = True, 
                    host_pid = True, 
                    hostname = '0', 
                    image_pull_secrets = [
                        None
                        ], 
                    init_containers = [
                        None
                        ], 
                    logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode = '0', 
                        url = '0', ), 
                    max_replicas = 56, 
                    min_replicas = 56, 
                    node_name = '0', 
                    node_selector = {
                        'key' : '0'
                        }, 
                    overhead = {
                        'key' : None
                        }, 
                    preemption_policy = '0', 
                    priority = 56, 
                    priority_class_name = '0', 
                    readiness_gates = [
                        None
                        ], 
                    restart_policy = '0', 
                    runtime_class_name = '0', 
                    scheduler_name = '0', 
                    security_context = None, 
                    service_account = '0', 
                    service_account_name = '0', 
                    share_process_namespace = True, 
                    subdomain = '0', 
                    termination_grace_period_seconds = 56, 
                    timeout = 56, 
                    tolerations = [
                        None
                        ], 
                    topology_spread_constraints = [
                        None
                        ], 
                    volumes = [
                        None
                        ], )
            )
        else :
            return V1beta1InferenceServiceSpec(
                predictor = kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                    active_deadline_seconds = 56, 
                    affinity = None, 
                    automount_service_account_token = True, 
                    batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                        max_batch_size = 56, 
                        max_latency = 56, 
                        timeout = 56, ), 
                    canary_traffic_percent = 56, 
                    container_concurrency = 56, 
                    containers = [
                        None
                        ], 
                    dns_config = None, 
                    dns_policy = '0', 
                    enable_service_links = True, 
                    ephemeral_containers = [
                        None
                        ], 
                    host_aliases = [
                        None
                        ], 
                    host_ipc = True, 
                    host_network = True, 
                    host_pid = True, 
                    hostname = '0', 
                    image_pull_secrets = [
                        None
                        ], 
                    init_containers = [
                        None
                        ], 
                    logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                        mode = '0', 
                        url = '0', ), 
                    max_replicas = 56, 
                    min_replicas = 56, 
                    node_name = '0', 
                    node_selector = {
                        'key' : '0'
                        }, 
                    onnx = kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                        args = [
                            '0'
                            ], 
                        command = [
                            '0'
                            ], 
                        env = [
                            None
                            ], 
                        env_from = [
                            None
                            ], 
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        ports = [
                            None
                            ], 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        volume_devices = [
                            None
                            ], 
                        volume_mounts = [
                            None
                            ], 
                        working_dir = '0', ), 
                    overhead = {
                        'key' : None
                        }, 
                    preemption_policy = '0', 
                    priority = 56, 
                    priority_class_name = '0', 
                    pytorch = kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0',
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    readiness_gates = [
                        None
                        ], 
                    restart_policy = '0', 
                    runtime_class_name = '0', 
                    scheduler_name = '0', 
                    security_context = None, 
                    service_account = '0', 
                    service_account_name = '0', 
                    share_process_namespace = True, 
                    sklearn = kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    subdomain = '0', 
                    tensorflow = kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    termination_grace_period_seconds = 56, 
                    timeout = 56, 
                    tolerations = [
                        None
                        ], 
                    topology_spread_constraints = [
                        None
                        ], 
                    triton = kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), 
                    volumes = [
                        None
                        ], 
                    xgboost = kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                        image = '0', 
                        image_pull_policy = '0', 
                        lifecycle = None, 
                        liveness_probe = None, 
                        name = '0', 
                        readiness_probe = None, 
                        resources = None, 
                        runtime_version = '0', 
                        security_context = None, 
                        startup_probe = None, 
                        stdin = True, 
                        stdin_once = True, 
                        storage_uri = '0', 
                        termination_message_path = '0', 
                        termination_message_policy = '0', 
                        tty = True, 
                        working_dir = '0', ), ),
        )

    def testV1beta1InferenceServiceSpec(self):
        """Test V1beta1InferenceServiceSpec"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Awaitable, Callable

from grpc import HandlerCallDetails, RpcMethodHandler
from grpc.aio import ServerInterceptor


class LoggingInterceptor(ServerInterceptor):

    async def intercept_service(
        self,
        continuation: Callable[[HandlerCallDetails], Awaitable[RpcMethodHandler]],
        handler_call_details: HandlerCallDetails,
    ) -> RpcMethodHandler:
        logging.info(f"grpc method: {handler_call_details.method}")
        return await continuation(handler_call_details)



# Copyright 2021 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# coding: utf-8

"""
    KServe

    Python SDK for KServe  # noqa: E501

    The version of the OpenAPI document: v0.1
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

import unittest
import datetime

import kserve
from kserve.models.v1beta1_inference_service_list import V1beta1InferenceServiceList  # noqa: E501
from kserve.rest import ApiException

class TestV1beta1InferenceServiceList(unittest.TestCase):
    """V1beta1InferenceServiceList unit test stubs"""

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def make_instance(self, include_optional):
        """Test V1beta1InferenceServiceList
            include_option is a boolean, when False only required
            params are included, when True both required and
            optional params are included """
        # model = kserve.models.v1beta1_inference_service_list.V1beta1InferenceServiceList()  # noqa: E501
        if include_optional :
            return V1beta1InferenceServiceList(
                api_version = '0', 
                items = [
                    kserve.models.v1beta1_inference_service.V1beta1InferenceService(
                        api_version = '0', 
                        kind = '0', 
                        metadata = None, 
                        spec = kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                            explainer = kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                aix = kserve.models.v1beta1_aix_explainer_spec.V1beta1AIXExplainerSpec(
                                    args = [
                                        '0'
                                        ], 
                                    command = [
                                        '0'
                                        ], 
                                    config = {
                                        'key' : '0'
                                        }, 
                                    env = [
                                        None
                                        ], 
                                    env_from = [
                                        None
                                        ], 
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    ports = [
                                        None
                                        ], 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    type = '0', 
                                    volume_devices = [
                                        None
                                        ], 
                                    volume_mounts = [
                                        None
                                        ], 
                                    working_dir = '0', ), 
                                alibi = kserve.models.v1beta1_alibi_explainer_spec.V1beta1AlibiExplainerSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    type = '0', 
                                    working_dir = '0', ), 
                                automount_service_account_token = True, 
                                batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                                    max_batch_size = 56, 
                                    max_latency = 56, 
                                    timeout = 56, ), 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                containers = [
                                    None
                                    ], 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                ephemeral_containers = [
                                    None
                                    ], 
                                host_aliases = [
                                    None
                                    ], 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                image_pull_secrets = [
                                    None
                                    ], 
                                init_containers = [
                                    None
                                    ], 
                                logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                                    mode = '0', 
                                    url = '0', ), 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                node_selector = {
                                    'key' : '0'
                                    }, 
                                overhead = {
                                    'key' : None
                                    }, 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                readiness_gates = [
                                    None
                                    ], 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                subdomain = '0', 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, 
                                tolerations = [
                                    None
                                    ], 
                                topology_spread_constraints = [
                                    None
                                    ], 
                                volumes = [
                                    None
                                    ], ), 
                            predictor = kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                automount_service_account_token = True, 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                onnx = kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                pytorch = kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0',
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                sklearn = kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                subdomain = '0', 
                                tensorflow = kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, 
                                triton = kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                xgboost = kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), ), 
                            transformer = kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                automount_service_account_token = True, 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                subdomain = '0', 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, ), ), 
                        status = kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                            address = None, 
                            components = {
                                'key' : kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                                    address = None, 
                                    latest_created_revision = '0', 
                                    latest_ready_revision = '0', 
                                    latest_rolledout_revision = '0',
                                    previous_rolledout_revision = '0',
                                    traffic = list(),
                                    url = None, )
                                }, 
                            conditions = [
                                None
                                ], 
                            observed_generation = 56, 
                            url = None, ), )
                    ], 
                kind = '0', 
                metadata = None
            )
        else :
            return V1beta1InferenceServiceList(
                items = [
                    kserve.models.v1beta1_inference_service.V1beta1InferenceService(
                        api_version = '0', 
                        kind = '0', 
                        metadata = None, 
                        spec = kserve.models.v1beta1_inference_service_spec.V1beta1InferenceServiceSpec(
                            explainer = kserve.models.v1beta1_explainer_spec.V1beta1ExplainerSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                aix = kserve.models.v1beta1_aix_explainer_spec.V1beta1AIXExplainerSpec(
                                    args = [
                                        '0'
                                        ], 
                                    command = [
                                        '0'
                                        ], 
                                    config = {
                                        'key' : '0'
                                        }, 
                                    env = [
                                        None
                                        ], 
                                    env_from = [
                                        None
                                        ], 
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    ports = [
                                        None
                                        ], 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    type = '0', 
                                    volume_devices = [
                                        None
                                        ], 
                                    volume_mounts = [
                                        None
                                        ], 
                                    working_dir = '0', ), 
                                alibi = kserve.models.v1beta1_alibi_explainer_spec.V1beta1AlibiExplainerSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    type = '0', 
                                    working_dir = '0', ), 
                                automount_service_account_token = True, 
                                batcher = kserve.models.v1beta1_batcher.V1beta1Batcher(
                                    max_batch_size = 56, 
                                    max_latency = 56, 
                                    timeout = 56, ), 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                containers = [
                                    None
                                    ], 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                ephemeral_containers = [
                                    None
                                    ], 
                                host_aliases = [
                                    None
                                    ], 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                image_pull_secrets = [
                                    None
                                    ], 
                                init_containers = [
                                    None
                                    ], 
                                logger = kserve.models.v1beta1_logger_spec.V1beta1LoggerSpec(
                                    mode = '0', 
                                    url = '0', ), 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                node_selector = {
                                    'key' : '0'
                                    }, 
                                overhead = {
                                    'key' : None
                                    }, 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                readiness_gates = [
                                    None
                                    ], 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                subdomain = '0', 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, 
                                tolerations = [
                                    None
                                    ], 
                                topology_spread_constraints = [
                                    None
                                    ], 
                                volumes = [
                                    None
                                    ], ), 
                            predictor = kserve.models.v1beta1_predictor_spec.V1beta1PredictorSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                automount_service_account_token = True, 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                onnx = kserve.models.v1beta1_onnx_runtime_spec.V1beta1ONNXRuntimeSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                pytorch = kserve.models.v1beta1_torch_serve_spec.V1beta1TorchServeSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0',
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                sklearn = kserve.models.v1beta1_sk_learn_spec.V1beta1SKLearnSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                subdomain = '0', 
                                tensorflow = kserve.models.v1beta1_tf_serving_spec.V1beta1TFServingSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, 
                                triton = kserve.models.v1beta1_triton_spec.V1beta1TritonSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), 
                                xgboost = kserve.models.v1beta1_xg_boost_spec.V1beta1XGBoostSpec(
                                    image = '0', 
                                    image_pull_policy = '0', 
                                    lifecycle = None, 
                                    liveness_probe = None, 
                                    name = '0', 
                                    readiness_probe = None, 
                                    resources = None, 
                                    runtime_version = '0', 
                                    security_context = None, 
                                    startup_probe = None, 
                                    stdin = True, 
                                    stdin_once = True, 
                                    storage_uri = '0', 
                                    termination_message_path = '0', 
                                    termination_message_policy = '0', 
                                    tty = True, 
                                    working_dir = '0', ), ), 
                            transformer = kserve.models.v1beta1_transformer_spec.V1beta1TransformerSpec(
                                active_deadline_seconds = 56, 
                                affinity = None, 
                                automount_service_account_token = True, 
                                canary_traffic_percent = 56, 
                                container_concurrency = 56, 
                                dns_config = None, 
                                dns_policy = '0', 
                                enable_service_links = True, 
                                host_ipc = True, 
                                host_network = True, 
                                host_pid = True, 
                                hostname = '0', 
                                max_replicas = 56, 
                                min_replicas = 56, 
                                node_name = '0', 
                                preemption_policy = '0', 
                                priority = 56, 
                                priority_class_name = '0', 
                                restart_policy = '0', 
                                runtime_class_name = '0', 
                                scheduler_name = '0', 
                                security_context = None, 
                                service_account = '0', 
                                service_account_name = '0', 
                                share_process_namespace = True, 
                                subdomain = '0', 
                                termination_grace_period_seconds = 56, 
                                timeout = 56, ), ), 
                        status = kserve.models.v1beta1_inference_service_status.V1beta1InferenceServiceStatus(
                            address = None, 
                            components = {
                                'key' : kserve.models.v1beta1_component_status_spec.V1beta1ComponentStatusSpec(
                                    address = None, 
                                    latest_created_revision = '0', 
                                    latest_ready_revision = '0',
                                    latest_rolledout_revision = '0',
                                    previous_rolledout_revision = '0',
                                    traffic = list(),
                                    url = None, )
                                }, 
                            conditions = [
                                None
                                ], 
                            observed_generation = 56, 
                            url = None, ), )
                    ],
        )

    def testV1beta1InferenceServiceList(self):
        """Test V1beta1InferenceServiceList"""
        inst_req_only = self.make_instance(include_optional=False)
        inst_req_and_optional = self.make_instance(include_optional=True)


if __name__ == '__main__':
    unittest.main()



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Union, Tuple, Optional

import cloudevents.exceptions as ce
import orjson
import pkg_resources
from cloudevents.http import CloudEvent, from_http
from cloudevents.sdk.converters.util import has_binary_headers
from ray.serve.api import RayServeHandle

from ..model import Model
from ..errors import InvalidInput, ModelNotFound
from ..model import ModelType
from ..model_repository import ModelRepository
from ..utils.utils import create_response_cloudevent
from .infer_type import InferRequest
from ..constants import constants
from .grpc import grpc_predict_v2_pb2 as pb
import time
import logging


class DataPlane:
    """KServe DataPlane
    """

    def __init__(self, model_registry: ModelRepository):
        self._model_registry = model_registry
        self._server_name = constants.KSERVE_MODEL_SERVER_NAME

        # Dynamically fetching version of the installed 'kserve' distribution. The assumption is
        # that 'kserve' will already be installed by the time this class is instantiated.
        self._server_version = pkg_resources.get_distribution("kserve").version

    @property
    def model_registry(self):
        return self._model_registry

    def get_model_from_registry(self, name: str) -> Union[Model, RayServeHandle]:
        model = self._model_registry.get_model(name)
        if model is None:
            raise ModelNotFound(name)

        return model

    def get_model(self, name: str) -> Union[Model, RayServeHandle]:
        """Get the model instance with the given name.

        The instance can be either ``Model`` or ``RayServeHandle``.

        Args:
            name (str): Model name.

        Returns:
            Model|RayServeHandle: Instance of the model.
        """
        model = self._model_registry.get_model(name)
        if model is None:
            raise ModelNotFound(name)
        if not self._model_registry.is_model_ready(name):
            model.load()
        return model

    @staticmethod
    def get_binary_cloudevent(body: Union[str, bytes, None], headers: Dict[str, str]) -> CloudEvent:
        """Helper function to parse CloudEvent body and headers.

        Args:
            body (str|bytes|None): Request body.
            headers (Dict[str, str]): Request headers.

        Returns:
            CloudEvent: A CloudEvent instance parsed from http body and headers.

        Raises:
            InvalidInput: An error when CloudEvent failed to parse.
        """
        try:
            # Use default unmarshaller if contenttype is set in header
            if "ce-contenttype" in headers:
                event = from_http(headers, body)
            else:
                event = from_http(headers, body, lambda x: x)

            return event
        except (ce.MissingRequiredFields, ce.InvalidRequiredFields, ce.InvalidStructuredJSON,
                ce.InvalidHeadersFormat, ce.DataMarshallerError, ce.DataUnmarshallerError) as e:
            raise InvalidInput(f"Cloud Event Exceptions: {e}")

    @staticmethod
    async def live() -> Dict[str, str]:
        """Server live.

        Returns ``{"status": "alive"}`` on successful invocation.
        Primarily meant to be used for Kubernetes liveness check.

        Returns:
            Dict: {"status": "alive"}
        """
        return {"status": "alive"}

    def metadata(self) -> Dict:
        """Server metadata.

        Note:
            Supports ``model_repository_extension`` as defined at Triton Server `Model Repository Extension`_.

        Returns:
            Returns a dict object with following fields:
                - name (str): name of the server.
                - version (str): server version number.
                - extension (list[str]): list of protocol extensions supported by this server.

        .. _Model Repository Extension:
            https://github.com/triton-inference-server/server/blob/main/docs/protocol/extension_model_repository.md
        """
        return {
            "name": self._server_name,
            "version": self._server_version,
            "extensions": ["model_repository_extension"]
        }

    async def model_metadata(self, model_name: str) -> Dict:
        """Get metadata for a specific model.

        Args:
            model_name (str): Model name

        Returns:
            Dict: dictionary with following fields:

                - name (str): name of the model
                - platform: "" (Empty String)
                - inputs: Dict with below fields
                    - name (str): name of the input
                    - datatype (str): Eg. INT32, FP32
                    - shape ([]int): The shape of the tensor.
                                   Variable-size dimensions are specified as -1.
                - outputs: Same as inputs described above.

        Note:
            For more information, check KServe v2 `Model Metadata`_ documentation.

        Raises:
            ModelNotFound: exception will be raised if the model with model_name is not found.

        .. _Model Metadata:
            https://github.com/kserve/kserve/blob/master/docs/predict-api/v2/required_api.md#model-metadata
        """
        # TODO: model versioning is not supported yet
        model = self.get_model_from_registry(model_name)

        if not isinstance(model, RayServeHandle):
            input_types = model.get_input_types()
            output_types = model.get_output_types()
        else:
            model_handle: RayServeHandle = model
            input_types = await model_handle.get_input_types.remote()
            output_types = await model_handle.get_output_types.remote()
        return {
            "name": model_name,
            "platform": "",
            "inputs": input_types,
            "outputs": output_types
        }

    @staticmethod
    async def ready() -> bool:
        """Server ready.

        Returns ``True``. Primarily meant to be used as Kubernetes readiness check.

        Returns:
            bool: True
        """
        return True

    def model_ready(self, model_name: str) -> bool:
        """Check if a model is ready.

        Args:
            model_name (str): name of the model

        Returns:
            bool: True if the model is ready, False otherwise.

        Raises:
            ModelNotFound: exception if model is not found
        """
        if self._model_registry.get_model(model_name) is None:
            raise ModelNotFound(model_name)

        return self._model_registry.is_model_ready(model_name)

    def decode(self, body, headers) -> Union[Dict, InferRequest]:
        t1 = time.time()
        if isinstance(body, InferRequest):
            return body
        if headers and has_binary_headers(headers):
            body = self.get_binary_cloudevent(body, headers)
        else:
            if type(body) is bytes:
                try:
                    body = orjson.loads(body)
                except orjson.JSONDecodeError as e:
                    raise InvalidInput(f"Unrecognized request format: {e}")
        t2 = time.time()
        logging.debug(f"decoded request in {round((t2 - t1) * 1000, 9)}ms")
        return body

    def encode(self, model_name, body, response, headers) -> Tuple[Dict, Dict[str, str]]:
        response_headers = {}
        # if we received a cloudevent, then also return a cloudevent
        is_cloudevent = False
        is_binary_cloudevent = False
        if headers:
            if has_binary_headers(headers):
                is_cloudevent = True
                is_binary_cloudevent = True
            if headers.get("content-type", "") == "application/cloudevents+json":
                is_cloudevent = True
        if is_cloudevent:
            response_headers, response = create_response_cloudevent(model_name, body, response,
                                                                    is_binary_cloudevent)

            if is_binary_cloudevent:
                response_headers["content-type"] = "application/json"
            else:
                response_headers["content-type"] = "application/cloudevents+json"
        return response, response_headers

    async def infer(
            self,
            model_name: str,
            body: Union[bytes, Dict, InferRequest],
            headers: Optional[Dict[str, str]] = None
    ) -> Tuple[Union[str, bytes, Dict, pb.ModelInferResponse], Dict[str, str]]:
        """Performs inference on the specified model with the provided body and headers.

        If the ``body`` contains an encoded `CloudEvent`_, then it will be decoded and processed.
        The response body/headers will also be encoded as CloudEvents.

        Args:
            model_name (str): Model name.
            body (bytes|Dict): Request body data.
            headers: (Optional[Dict[str, str]]): Request headers.

        Returns:
            Tuple[Union[str, bytes, Dict], Dict[str, str]]:
                - response: The inference result.
                - response_headers: Headers to construct the HTTP response.

        Raises:
            InvalidInput: An error when the body bytes can't be decoded as JSON.

        .. _CloudEvent: https://cloudevents.io/
        """
        body = self.decode(body, headers)

        # call model locally or remote model workers
        model = self.get_model(model_name)
        if not isinstance(model, RayServeHandle):
            response = await model(body, headers=headers)
        else:
            model_handle: RayServeHandle = model
            response = await model_handle.remote(body)

        response, response_headers = self.encode(model_name, body, response, headers)
        return response, response_headers

    async def explain(self, model_name: str,
                      body: Union[bytes, Dict, InferRequest],
                      headers: Optional[Dict[str, str]] = None
                      ) -> Tuple[Union[str, bytes, Dict], Dict[str, str]]:
        """Performs explanation for the specified model.

        Args:
            model_name (str): Model name to be used for explanation.
            body (bytes|Dict): Request body data.
            headers: (Optional[Dict[str, str]]): Request headers.

        Returns:
            Dict: Explanation result.

        Raises:
            InvalidInput: An error when the body bytes can't be decoded as JSON.
        """
        body = self.decode(body, headers)

        # call model locally or remote model workers
        model = self.get_model(model_name)
        if not isinstance(model, RayServeHandle):
            response = await model(body, model_type=ModelType.EXPLAINER)
        else:
            model_handle = model
            response = await model_handle.remote(body, model_type=ModelType.EXPLAINER)
        response, response_headers = self.encode(model_name, body, response, headers)
        return response, response_headers



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import logging.config

from .constants.constants import KSERVE_LOGLEVEL

KSERVE_LOGGER_NAME = 'kserve'
KSERVE_TRACE_LOGGER_NAME = 'kserve.trace'
KSERVE_LOGGER_FORMAT = ('%(asctime)s.%(msecs)03d %(process)s %(name)s '
                        '%(levelname)s [%(funcName)s():%(lineno)s] %(message)s')
KSERVE_TRACE_LOGGER_FORMAT = ('%(asctime)s.%(msecs)03d %(name)s %(message)s')
KSERVE_LOGGER_DATE_FORMAT = "%Y-%m-%d %H:%M:%S"


KSERVE_LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "kserve": {
            "()": "logging.Formatter",
            "fmt": KSERVE_LOGGER_FORMAT,
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
        },
        "kserve_trace": {
            "()": "logging.Formatter",
            "fmt": KSERVE_TRACE_LOGGER_FORMAT,
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
        },
        "uvicorn": {
            "()": "uvicorn.logging.DefaultFormatter",
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
            "fmt": "%(asctime)s.%(msecs)03d %(name)s %(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "uvicorn_access": {
            "()": "uvicorn.logging.AccessFormatter",
            "datefmt": KSERVE_LOGGER_DATE_FORMAT,
            "fmt": '%(asctime)s.%(msecs)03d %(name)s '
                   '%(levelprefix)s %(client_addr)s %(process)s - '
                   '"%(request_line)s" %(status_code)s',
            # noqa: E501
        },
    },
    "handlers": {
        "kserve": {
            "formatter": "kserve",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "kserve_trace": {
            "formatter": "kserve_trace",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "uvicorn": {
            "formatter": "uvicorn",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "uvicorn_access": {
            "formatter": "uvicorn_access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
    },
    "loggers": {
        "kserve": {"handlers": ["kserve"], "level": KSERVE_LOGLEVEL, "propagate": False},
        "kserve.trace": {"handlers": ["kserve_trace"],
                         "level": KSERVE_LOGLEVEL, "propagate": False},
        "uvicorn": {"handlers": ["uvicorn"], "level": "INFO", "propagate": False},
        "uvicorn.error": {"level": "INFO"},
        "uvicorn.access": {
            "handlers": ["uvicorn_access"],
            "level": "INFO",
            "propagate": False
        },
    },
}

logger = logging.getLogger(KSERVE_LOGGER_NAME)
trace_logger = logging.getLogger(KSERVE_TRACE_LOGGER_NAME)



# Copyright 2022 The KServe Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import io
import json
import os
import pathlib
import re
from unittest import mock

import avro
import pytest
from cloudevents.conversion import to_binary, to_structured
from cloudevents.http import CloudEvent
from ray import serve

from kserve.errors import InvalidInput, ModelNotFound
from kserve.protocol.dataplane import DataPlane
from kserve.model_repository import ModelRepository
from test.test_server import DummyModel, dummy_cloud_event, DummyCEModel, DummyAvroCEModel, \
    DummyServeModel


@pytest.mark.asyncio
class TestDataPlane:
    MODEL_NAME = "TestModel"

    # Adding two 'params' to the below fixture causes all the tests using this fixture to run twice.
    # First time this fixture uses 'Model' class.
    # The second time it uses the 'RayServeHandle' class.
    @pytest.fixture(scope="class", params=["TEST_RAW_MODEL", "TEST_RAY_SERVE_MODEL"])
    def dataplane_with_model(self, request):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())

        if request.param == "TEST_RAW_MODEL":
            model = DummyModel(self.MODEL_NAME)
            model.load()
            dataplane._model_registry.update(model)
            yield dataplane
        else:  # request.param == "TEST_RAY_SERVE_MODEL"
            serve.start(detached=False, http_options={"host": "0.0.0.0", "port": 9071})
            DummyServeModel.deploy(self.MODEL_NAME)
            handle = DummyServeModel.get_handle()
            handle.load.remote()
            dataplane._model_registry.update_handle(self.MODEL_NAME, handle)
            yield dataplane
            serve.shutdown()

    async def test_get_model_from_registry(self):
        dataplane = DataPlane(model_registry=ModelRepository())
        model_name = "FakeModel"
        with pytest.raises(ModelNotFound) as http_exec:
            dataplane.get_model_from_registry(model_name)
        # assert http_exec.value.status_code == 404
        assert http_exec.value.reason == f"Model with name {model_name} does not exist."

        ready_model = DummyModel("Model")
        ready_model.load()
        dataplane._model_registry.update(ready_model)
        dataplane.get_model_from_registry("Model")

    async def test_liveness(self):
        assert (await DataPlane.live()) == {'status': 'alive'}

    async def test_server_readiness(self, dataplane_with_model):
        assert (await dataplane_with_model.ready()) is True

        not_ready_model = DummyModel("NotReadyModel")
        # model.load()  # Model not loaded, i.e. model not ready
        dataplane_with_model._model_registry.update(not_ready_model)
        # The model server readiness endpoint should return 'True' irrespective of the readiness
        # of the model loaded into it.
        assert (await dataplane_with_model.ready()) is True

    async def test_model_readiness(self):
        dataplane = DataPlane(model_registry=ModelRepository())

        ready_model = DummyModel("ReadyModel")
        ready_model.load()
        dataplane._model_registry.update(ready_model)
        assert dataplane.model_ready(ready_model.name) is True

        not_ready_model = DummyModel("NotReadyModel")
        # model.load()  # Model not loaded, i.e. not ready
        dataplane._model_registry.update(not_ready_model)
        assert dataplane.model_ready(not_ready_model.name) is False

    async def test_server_metadata(self):
        with open(pathlib.Path(__file__).parent.parent.parent / 'VERSION') as version_file:
            version = version_file.read().strip()

        dataplane = DataPlane(model_registry=ModelRepository())
        expected_metadata = {
            "name": "kserve",
            "version": version,
            "extensions": ["model_repository_extension"]
        }
        assert dataplane.metadata() == expected_metadata

    async def test_model_metadata(self, dataplane_with_model):
        assert await dataplane_with_model.model_metadata(self.MODEL_NAME) == {
            "name": self.MODEL_NAME,
            "platform": "",
            "inputs": [],
            "outputs": []
        }

    async def test_infer(self, dataplane_with_model):
        body = b'{"instances":[[1,2]]}'
        resp = await dataplane_with_model.infer(self.MODEL_NAME, body)
        assert resp == (
            {"predictions": [[1, 2]]},  # body
            {}  # headers
        )

    async def test_explain(self, dataplane_with_model):
        body = b'{"instances":[[1,2]]}'
        resp = await dataplane_with_model.explain(self.MODEL_NAME, body)

        assert resp == (
            {"predictions": [[1, 2]]},
            {}
        )


@pytest.mark.asyncio
class TestDataPlaneCloudEvent:
    MODEL_NAME = "TestModel"

    @pytest.fixture(scope="class")
    def dataplane_with_ce_model(self):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())
        model = DummyCEModel(self.MODEL_NAME)
        model.load()
        dataplane._model_registry.update(model)
        return dataplane

    async def test_infer_ce_structured(self, dataplane_with_ce_model):
        event: CloudEvent = dummy_cloud_event({"instances": [[1, 2]]})
        headers, body = to_structured(event)

        resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
        body = json.loads(resp[0])
        headers = resp[1]

        assert headers['content-type'] == "application/cloudevents+json"

        assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert body["data"] == {"predictions": [[1, 2]]}
        assert body['specversion'] == "1.0"
        assert body['source'] == "io.kserve.inference.TestModel"
        assert body['type'] == "io.kserve.inference.response"
        assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    async def test_infer_custom_ce_attributes(self, dataplane_with_ce_model):
        with mock.patch.dict(
                os.environ,
                {
                    "CE_SOURCE": "io.kserve.inference.CustomSource",
                    "CE_TYPE": "io.kserve.custom_type"
                }
        ):
            event = dummy_cloud_event({"instances": [[1, 2]]})
            headers, body = to_structured(event)

            resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
            body = json.loads(resp[0])
            headers = resp[1]

            assert headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.inference.CustomSource"
            assert body['type'] == "io.kserve.custom_type"

    async def test_infer_merge_structured_ce_attributes(self, dataplane_with_ce_model):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, add_extension=True)
            headers, body = to_structured(event)

            resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
            body = json.loads(resp[0])
            headers = resp[1]

            assert headers['content-type'] == "application/cloudevents+json"

            assert body["id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            assert body["data"] == {"predictions": [[1, 2]]}
            assert body['source'] == "io.kserve.inference.TestModel"
            assert body['type'] == "io.kserve.inference.response"
            assert body[
                       "custom-extension"] == "custom-value"  # Added by add_extension=True in dummy_cloud_event
            assert body['time'] > "2021-01-28T21:04:43.144141+00:00"

    async def test_infer_merge_binary_ce_attributes(self, dataplane_with_ce_model):
        with mock.patch.dict(os.environ, {"CE_MERGE": "true"}):
            event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True,
                                      add_extension=True)
            headers, body = to_binary(event)

            resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
            body = resp[0]
            headers = resp[1]

            assert headers['content-type'] == "application/json"
            assert headers['ce-specversion'] == "1.0"
            assert headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
            # Added by add_extension=True in dummy_cloud_event
            assert headers['ce-custom-extension'] == 'custom-value'
            assert headers['ce-source'] == "io.kserve.inference.TestModel"
            assert headers['ce-type'] == "io.kserve.inference.response"
            assert headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
            assert body == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_binary_dict(self, dataplane_with_ce_model):
        event = dummy_cloud_event({"instances": [[1, 2]]}, set_contenttype=True)
        headers, body = to_binary(event)

        resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
        body = resp[0]
        headers = resp[1]

        assert headers['content-type'] == "application/json"
        assert headers['ce-specversion'] == "1.0"
        assert headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert headers['ce-source'] == "io.kserve.inference.TestModel"
        assert headers['ce-type'] == "io.kserve.inference.response"
        assert headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert body == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_binary_bytes(self, dataplane_with_ce_model):
        event = dummy_cloud_event(b'{"instances":[[1,2]]}', set_contenttype=True)
        headers, body = to_binary(event)

        resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
        body = resp[0]
        headers = resp[1]

        assert headers['content-type'] == "application/json"
        assert headers['ce-specversion'] == "1.0"
        assert headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert headers['ce-source'] == "io.kserve.inference.TestModel"
        assert headers['ce-type'] == "io.kserve.inference.response"
        assert headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert body == b'{"predictions": [[1, 2]]}'

    async def test_infer_ce_bytes_bad_format_exception(self, dataplane_with_ce_model):
        event = dummy_cloud_event(b'{', set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(InvalidInput) as err:
            await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)

        error_regex = re.compile("Failed to decode or parse binary json cloudevent: "
                                 "unexpected end of data:*")
        assert error_regex.match(err.value.reason) is not None

    async def test_infer_ce_bytes_bad_hex_format_exception(self, dataplane_with_ce_model):
        event = dummy_cloud_event(b'0\x80\x80\x06World!\x00\x00', set_contenttype=True)
        headers, body = to_binary(event)

        with pytest.raises(InvalidInput) as err:
            await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)

        error_regex = re.compile("Failed to decode or parse binary json cloudevent: 'utf-8' codec "
                                 "can't decode byte 0x80 in position 1: invalid start byte.*")

        assert error_regex.match(err.value.reason) is not None


@pytest.mark.asyncio
class TestDataPlaneAvroCloudEvent:
    test_avsc_schema = '''
            {
            "namespace": "example.avro",
             "type": "record",
             "name": "User",
             "fields": [
                 {"name": "name", "type": "string"},
                 {"name": "favorite_number",  "type": ["int", "null"]},
                 {"name": "favorite_color", "type": ["string", "null"]}
             ]
            }
            '''

    MODEL_NAME = "TestModel"

    @pytest.fixture(scope="class")
    def dataplane_with_ce_model(self):  # pylint: disable=no-self-use
        dataplane = DataPlane(model_registry=ModelRepository())
        model = DummyAvroCEModel(self.MODEL_NAME)
        model.load()
        dataplane._model_registry.update(model)
        return dataplane

    async def test_infer_ce_avro_binary(self, dataplane_with_ce_model):
        schema = avro.schema.parse(self.test_avsc_schema)
        msg = {"name": "foo", "favorite_number": 1, "favorite_color": "pink"}

        writer = avro.io.DatumWriter(schema)
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer.write(msg, encoder)
        data = bytes_writer.getvalue()

        event = dummy_cloud_event(data, set_contenttype=True)
        # Creates the HTTP request representation of the CloudEvent in binary content mode
        headers, body = to_binary(event)

        resp = await dataplane_with_ce_model.infer(self.MODEL_NAME, body, headers)
        body = resp[0]
        headers = resp[1]

        assert headers['content-type'] == "application/json"
        assert headers['ce-specversion'] == "1.0"
        assert headers["ce-id"] != "36077800-0c23-4f38-a0b4-01f4369f670a"
        assert headers['ce-source'] == "io.kserve.inference.TestModel"
        assert headers['ce-type'] == "io.kserve.inference.response"
        assert headers['ce-time'] > "2021-01-28T21:04:43.144141+00:00"
        assert body == b'{"predictions": [["foo", 1, "pink"]]}'
